/*
 This code computes the Local Rank Distance between the strings (or texts)
 given in an input file.
 
 Copyright (C) 2015  Radu Tudor Ionescu
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or any
 later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.util.*;

public class ComputeDistanceMatrix
{
    private int ngramLength;
    private int maxOffset;
    
    public ComputeDistanceMatrix(int ngramLength, int maxOffset)
    {
        this.ngramLength = ngramLength;
        this.maxOffset = maxOffset;
    }

    public void makeDistanceMatrixForInputSamples(String inputFile, String outputFile)
    {
        String text;
        int n = 0;

        Vector<String> samples = new Vector<String>();

        BufferedReader in;

        try
        {
            in = new BufferedReader(new InputStreamReader(new FileInputStream(inputFile), "UTF-8"));
            text = in.readLine();
            
            while (text != null)
            {
               text = text.trim();
               text = text.toLowerCase();
               text = text.replaceAll("\\s+", " ");
               samples.add(text);
               n++;
               text = in.readLine();
            }
            in.close();

            System.out.println("Loaded " + n + " samples from " + inputFile);

        }
        catch (IOException e)
        {
            System.out.println(e);
            System.exit(1);
        }
        
        System.out.println("Computing the LRD matrix based on " + ngramLength + "-grams ...");

        
        LocalRankDistance LRD = new LocalRankDistance(ngramLength, maxOffset);
            
        float[][] D = LRD.computeLocalRankDistanceBetweenSamples(samples);
            
        System.out.println("Saving distance matrix to " + outputFile + " ...");
        writeDistanceMatrix(D, samples, outputFile, true);
    }

    public void writeDistanceMatrix(float[][] D, Vector<String> samples, String outputFile, boolean normalized)
    {
        BufferedWriter out;
        try
        {
            out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(outputFile)));
            
            for (int i = 0; i < D.length; i++)
            {
                for (int j = 0; j < D.length; j++)
                {
                    float dist = (normalized) ? D[i][j] / (maxOffset * (samples.get(i).length() + samples.get(j).length())) : D[i][j];
                    out.write(dist + " ");
                }
                out.write("\n");
            }
            out.close();
        }
        catch (IOException e)
        {
            System.out.println(e);
            System.exit(1);
        }
    }
    
    public static void main(String[] args)
    {
        if (args.length != 4)
        {
            System.out.println("Usage: java ComputeDistanceMatrix <ngramLength> <maxOffset> <inputFile> <outputFile>");
            System.out.println("Parameters:");
            System.out.println("\t<ngramMinLength> specifies the n-gram length;");
            System.out.println("\t<maxOffset> specifies the maximum spatial offset among identical n-grams;");
            System.out.println("\t<inputFile> is an input text file that contains a set of text samples (one sample per row);");
            System.out.println("\t<outputFile> is the generated output file with the pairwise distance matrix.");
        }
        else
        {
            ComputeDistanceMatrix CDM = new ComputeDistanceMatrix(Integer.parseInt(args[0]), Integer.parseInt(args[1]));
            CDM.makeDistanceMatrixForInputSamples(args[2], args[3]);
        }
    }
}
