/*
 This code computes the Local Rank Distance.
 
 Copyright (C) 2015  Radu Tudor Ionescu, Marius Popescu
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or any
 later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.util.*;

public class LocalRankDistance
{
    private Vector<String> samples;

    private int ngramSize;
    private int maxOffset;
    
    public LocalRankDistance(int ngramSize, int maxOffset)
    {
        this.ngramSize = ngramSize;
        this.maxOffset = maxOffset;
    }

    // Computes local rank distance between strings given an inverted index and a string index using an optimized implementation
    float hashLocalRankDistance(HashMap<String, ArrayList<Integer> > ngramsHash, int stringIndex)
    {
        int i, j;
        int minPos, midPos, maxPos;
        int minOffset, offset;
        
        float dist = 0f;
        String ngram;
        
        for (i = 0; i <= samples.get(stringIndex).length() - ngramSize; i++)
        {
            ngram = samples.get(stringIndex).substring(i, i + ngramSize);
            
            if (ngramsHash.containsKey(ngram))
            {
                minOffset = maxOffset;
                ArrayList<Integer> positions = ngramsHash.get(ngram);
                
                /*
                for(j = 0; j < positions.size(); j++)
                {
                 offset = (int)abs(positions[j] - i);
                 if (offset < minOffset) minOffset = offset;
                }
                */
                
                minPos = 0;
                maxPos = positions.size() - 1;
                
                while (maxPos - minPos > 1)
                {
                    midPos = (minPos + maxPos) >> 1;
                    if (i == positions.get(midPos))
                    {
                        minPos = midPos;
                        maxPos = midPos;
                    }
                    else if (i < positions.get(midPos)) maxPos = midPos;
                    else minPos = midPos;
                }
                for(j = minPos; j <= maxPos; j++)
                {
                    offset = (int)Math.abs(positions.get(j) - i);
                    if (offset < minOffset) minOffset = offset;
                }
                
                dist += minOffset;
            }
            else dist += maxOffset;
        }
        
        return dist;
    }
    
    float[][] computeLocalRankDistanceBetweenSamples(Vector<String> samples)
    {
        if (samples.size() < 0) return null;
        this.samples = samples;

        int i, j;
        String ngram;
        
        float meanStringSize = 0f;
        for (i = 0; i < samples.size(); i++)
            meanStringSize += samples.get(i).length();
        System.out.println("Mean string size : " + (meanStringSize / samples.size()));
        
        float[][] D = new float[samples.size()][samples.size()];
        for (i = 0; i < samples.size(); i++)
            for (j = 0; j < samples.size(); j++)
                D[i][j] = 0f;
        
        float localRankDist;
        for (i = 0; i < samples.size(); i++)
        {
            // System.out.println("Computing local rank distances for string " + i);
            
            // Build positional inverted index for the string i
            HashMap<String, ArrayList<Integer> > ngramsHash = new HashMap<String, ArrayList<Integer> >();
            
            for (j = 0; j <= samples.get(i).length() - ngramSize; j++)
            {
                ngram = samples.get(i).substring(j, j + ngramSize);
                ArrayList<Integer> positions = ngramsHash.get(ngram);
                
                if (positions != null) positions.add(j);
                else
                {
                    positions = new ArrayList<Integer>();
                    positions.add(j);
                }
                
                ngramsHash.put(ngram, positions);
            }
            
            for (j = 0; j < samples.size(); j++)
                if (i != j)
                {
                    localRankDist = hashLocalRankDistance(ngramsHash, j);
                    D[i][j] += localRankDist;
                    D[j][i] += localRankDist;
                }
            
            if ((i + 1) % 100 == 0) System.out.println("Computed distance matrix to index " + i);
        }
        
        return D;
    }
}
